#ifndef _TRACE_H
#define _TRACE_H

//#define TRACE_ON

enum trace_codes {
    TC_EXIT,
    TC_NDR_OVERHEAD,
    TC_NDR_KERNEL_START,
    TC_NDR_KERNEL_COMPLETE,
    TC_NDR_CACHE_COHERENCE_COMPLETE,
    TC_OOT_OVERHEAD,
    TC_OOT_KERNEL_START,
    TC_OOT_KERNEL_COMPLETE,
    TC_OOT_CACHE_COHERENCE_COMPLETE,
    TC_IOT_OVERHEAD,
    TC_IOT_KERNEL_START,
    TC_IOT_KERNEL_COMPLETE,
    TC_IOT_CACHE_COHERENCE_COMPLETE
};

static inline char *decode_tc(enum trace_codes tc)
{
    switch(tc)
    {
        case TC_EXIT:                           return "Monitor Exit";
        case TC_NDR_OVERHEAD:                   return "NDR Overhead";
        case TC_NDR_KERNEL_START:               return "NDR Start";
        case TC_NDR_KERNEL_COMPLETE:            return "NDR Complete";
        case TC_NDR_CACHE_COHERENCE_COMPLETE:   return "NDR Cache Complete";
        case TC_OOT_OVERHEAD:                   return "OOT Overhead";
        case TC_OOT_KERNEL_START:               return "OOT Start";
        case TC_OOT_KERNEL_COMPLETE:            return "OOT Complete";
        case TC_OOT_CACHE_COHERENCE_COMPLETE:   return "OOT Cache Complete";
        case TC_IOT_OVERHEAD:                   return "IOT Overhead";
        case TC_IOT_KERNEL_START:               return "IOT Start";
        case TC_IOT_KERNEL_COMPLETE:            return "IOT Complete";
        case TC_IOT_CACHE_COHERENCE_COMPLETE:   return "IOT Cache Complete";
    }
    return "UNKNOWN";
}

#if defined(TRACE_ON)
#define TRACE(code, kid, wgid) \
    if (code != TC_EXIT && \
        code != TC_NDR_CACHE_COHERENCE_COMPLETE && \
        code != TC_OOT_CACHE_COHERENCE_COMPLETE && \
        code != TC_IOT_CACHE_COHERENCE_COMPLETE) \
            printf("%1d %-18s %5d %5d %u\n", DNUM,decode_tc(code),kid,wgid,readClockGlobal()) 
#else
#define TRACE(code, kid, wgid) 
#endif

#if 0
static inline void stats_trace(enum trace_codes tc)
{
    switch(tc)
    {
        case TC_NDR_OVERHEAD:                   
        case TC_OOT_OVERHEAD:                   
        case TC_IOT_OVERHEAD:                   
            stats_start();
            break;

        case TC_NDR_KERNEL_START:               
        case TC_OOT_KERNEL_START:               
        case TC_IOT_KERNEL_START:               
            start_time = readClockGlobal();
            break;

        case TC_NDR_KERNEL_COMPLETE:            
        case TC_OOT_KERNEL_COMPLETE:            
        case TC_IOT_KERNEL_COMPLETE:            
            stop_time = readClockGlobal();
            stats_record(wgid, DNUM, start_time, stop_time);
            break;

        case TC_NDR_CACHE_COHERENCE_COMPLETE:   
            if (DNUM == 0) stats_stop();
            break;

        case TC_OOT_CACHE_COHERENCE_COMPLETE:   
        case TC_IOT_CACHE_COHERENCE_COMPLETE:   
        case TC_EXIT:                           
            break;
    }
}
#endif


#endif // _TRACE_H

